/*
 * Generic SOL Virtual Machine
 * By Nils M Holm, 2010
 * IN THE PUBLIC DOMAIN
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <curses.h>

#define MEMSIZE	262144
#define BLKSIZE	1024
#define STKLEN	1024

#define	byte	unsigned char
#define word	int
#define uword	unsigned word

#define GETW(x)		(  (Mem[x]<<24)		\
			 | (Mem[(x)+1]<<16)	\
			 | (Mem[(x)+2]<<8) 	\
			 |  Mem[(x)+3])

#define PUTW(x,v)	((Mem[x] = ((v)>>24)),		\
			 (Mem[(x)+1] = ((v)>>16)),	\
			 (Mem[(x)+2] = ((v)>>8)),	\
			 (Mem[(x)+3] = ((v))))

#define TOS()		GETW(Sp)
#define TOS2()		GETW(Sp+4)
#define RTOS()		GETW(Rp)

#define TO_TOS(x)	PUTW(Sp, x)
#define TO_TOS2(x)	PUTW(Sp+4, x)
#define TO_RTOS(x)	PUTW(Rp, x)

int	Image_fd;
byte	Mem[MEMSIZE];
word	Ip, Tp, Sp, Rp;
int	color;

#define ICALL	0x00
#define IHALT	0x80
#define IINTP	0x81
#define INEXT	0x82
#define ILIT	0x83
#define ICONST	0x84
#define IARRAY	0x85
#define IEXIT	0x86
#define IEXEC	0x87
#define IBR	0x88
#define IBF	0x89
#define IREDIR	0x8a
#define IDUP	0x90
#define IDROP	0x91
#define ISWAP	0x92
#define IOVER	0x93
#define IROT	0x94
#define IGET	0x95
#define ICGET	0x96
#define IPUT	0x97
#define ICPUT	0x98
#define ISPGET	0x99
#define ISPPUT	0x9a
#define IRPGET	0x9b
#define IRPPUT	0x9c
#define ITOR	0x9d
#define IFROMR	0x9e
#define IRGET	0x9f
#define IRDROP	0xa0
#define IPLUS	0xb0
#define IMINUS	0xb1
#define IMUL	0xb2
#define IUMUL	0xb3
#define IDIVMOD	0xb4
#define IUDIVMD	0xb5
#define IAND	0xb6
#define IOR	0xb7
#define IXOR	0xb8
#define ISHL	0xb9
#define ISHR	0xba
#define INEG	0xbb
#define IINV	0xbc
#define IINC	0xbd
#define IDEC	0xbe
#define IADDTO	0xbf
#define IBSZ	0xc0
#define INOT	0xc1
#define ILESS	0xc2
#define IGRTR	0xc3
#define IULESS	0xc4
#define IUGRTR	0xc5
#define IEQUAL	0xc6
#define ICMOVE	0xd0
#define ICSCAN	0xd1
#define IDTEXT	0xd2
#define IFILL	0xd3
#define ILOOKUP	0xd4
#define IKBRDY	0xe0
#define IKEY	0xe1
#define IEMIT	0xe2
#define ICURSOR	0xe3
#define ICOLOR	0xe4
#define ICPOS	0xe5
#define ICLREOL	0xe6
#define ICLRSCR	0xe7
#define IDSKIO	0xf0

void init(void) {
	int	colors[] = { COLOR_BLACK, COLOR_BLUE, COLOR_GREEN,
				COLOR_CYAN, COLOR_RED, COLOR_MAGENTA,
				COLOR_YELLOW|A_STANDOUT, COLOR_WHITE };
	int	f, b;

	initscr();
	start_color();
	for (b=0; b<8; b++) {
		for (f=0; f<8; f++) {
			init_pair(b*8+f, colors[f], colors[b]);
		}
	}
	raw();
	noecho();
	nonl();
	idlok(stdscr, 1);
	scrollok(stdscr, 1);
	keypad(stdscr, 1);
}

void fini(void) {
	close(Image_fd);
	refresh();
	endwin();
}

void quit(char *msg) {
	addstr(msg);
	fini();
	exit(2);
}

void load(void) {
	int	offset, length;

	clear();
	mvaddstr(0, 0, "Loading SOL: ");
	refresh();
	Image_fd = open("soldisk", O_RDWR);
	if (Image_fd < 0)
		quit("no disk image found (soldisk)");
	if (read(Image_fd, Mem, BLKSIZE) != BLKSIZE)
		quit("corrupt disk image");
	offset = GETW(0);
	length = GETW(4);
	lseek(Image_fd, offset * BLKSIZE, SEEK_SET);
	if (read(Image_fd, &Mem[BLKSIZE], BLKSIZE*length) != BLKSIZE*length)
		quit("corrupt disk image");
	if (Mem[BLKSIZE] != 0x80 || 0 != memcmp(&Mem[BLKSIZE+1], "SOL", 3))
		quit("corrupt disk image");
	printw("block %d, length %d\n\r", offset, length);
	refresh();
}

void printregs(char *msg) {
	printw("IP=%x SP=%x RP=%x (IP)=%x", Ip, Sp, Rp, Mem[Ip-1]);
	if (Sp != MEMSIZE)
		printw(" TOS=%x", TOS());
	if (Rp != MEMSIZE-STKLEN*4)
		printw(" RTOS=%x", RTOS());
	if (msg) {
		addstr(" - ");
		addstr(msg);
	}
	addstr("\n\r");
}

void push(word x) {
	Sp -= 4;
	PUTW(Sp,x);
}

word pop(void) {
	Sp += 4;
	return GETW(Sp-4);
}

void rpush(word x) {
	Rp -= 4;
	PUTW(Rp,x);
}

word rpop(void) {
	Rp += 4;
	return GETW(Rp-4);
}

void disk_io(void) {
	word	dev, blk, a, op, r;

	op = pop();
	a = pop();
	blk = pop();
	dev = pop();
/* printf("dev=%d blk=%d a=%x op=%d", dev, blk, a, op); */
	lseek(Image_fd, blk*BLKSIZE, SEEK_SET);
	switch (op) {
	case 0:
		r = read(Image_fd, &Mem[a], BLKSIZE);
		break;
	case 1:
		r = write(Image_fd, &Mem[a], BLKSIZE);
		break;
	default:
		r = -1;
		break;
	}
/* printf(" r=%d\n", r); */
	push(BLKSIZE == r? 0: -1);
}

void lookup(void) {
	word	lex, tok, len;
	byte	*s1, *name;

	lex = pop();
	tok = pop();
	name = &Mem[tok+1];
	len = Mem[tok];
	while (lex) {
		s1 = &Mem[lex+4];
		if (len == (s1[0] & 0x7f) &&
		    !memcmp(name, &s1[1], len))
		{
			push(lex);
			push(-1);
			return;
		}
		lex = GETW(lex);
	}
	push(0);
}

int getkey(void) {
	int c = getch();

	switch (c) {
	case KEY_UP:		return 256;
	case KEY_DOWN:		return 257;
	case KEY_LEFT:		return 258;
	case KEY_RIGHT:		return 259;
	case KEY_BACKSPACE:	return 8;
	default:		return c;
	}
}

void run() {
	word	i, j, k;
	uword	u, v, w;

	Sp = MEMSIZE - 2*BLKSIZE - 0x100;
	Rp = Sp - STKLEN*4;
	Ip = GETW(1028);
	Tp = 4;
	printregs("started");
	refresh();
	for (;;) {
		switch (Mem[Ip++]) {
		case ICALL:
			rpush(Ip+3);
			Ip = GETW(Ip-1);
			break;
		case IHALT:
			printregs("halted");
			fini();
			exit(0);
			break;
		case IINTP:
			i = RTOS();
			TO_RTOS(Tp);
			Tp = i;
			/* FALL THROUGH */
		case INEXT:
			Ip = GETW(Tp);
			Tp += 4;
			break;
		case ILIT:
			push(GETW(Tp));
			Tp += 4;
			break;
		case ICONST:
			i = rpop();
			push(GETW(i));
			break;
		case IARRAY:
			i = rpop();
			push(i+pop()*4);
			break;
		case IEXIT:
			Tp = rpop();
			break;
		case IEXEC:
			Ip = pop();
			break;
		case IBR:
			Tp = GETW(Tp);
			break;
		case IBF:
			if (0 == pop())
				Tp = GETW(Tp);
			else
				Tp += 4;
			break;
		case IREDIR:
			Ip = GETW(Ip);
			break;
		case IDUP:
			push(TOS());
			break;
		case IDROP:
			Sp += 4;
			break;
		case ISWAP:
			i = TOS();
			j = TOS2();
			TO_TOS(j);
			TO_TOS2(i);
			break;
		case IOVER:
			push(TOS2());
			break;
		case IROT:
			i = GETW(Sp+8);
			PUTW(Sp+8, TOS2());
			j = TOS();
			TO_TOS2(j);
			TO_TOS(i);
			break;
		case IGET:
			i = TOS();
			i = GETW(i);
			TO_TOS(i);
			break;
		case ICGET:
			i = TOS();
			i = Mem[i];
			TO_TOS(i);
			break;
		case IPUT:
			u = pop();
			i = pop();
			PUTW(u, i);
			break;
		case ICPUT:
			u = pop();
			v = pop();
			Mem[u] = v;
			break;
		case ISPGET:
			push(Sp);
			break;
		case ISPPUT:
			Sp = pop();
			break;
		case IRPGET:
			push(Rp);
			break;
		case IRPPUT:
			Rp = pop();
			break;
		case ITOR:
			rpush(pop());
			break;
		case IFROMR:
			push(rpop());
			break;
		case IRGET:
			push(RTOS());
			break;
		case IRDROP:
			rpop();
			break;
		case IPLUS:
			i = TOS() + TOS2();
			TO_TOS2(i);
			Sp += 4;
			break;
		case IMINUS:
			i = TOS2() - TOS();
			TO_TOS2(i);
			Sp += 4;
			break;
		case IMUL:
		case IUMUL:
			i = TOS() * TOS2();
			TO_TOS2(i);
			Sp += 4;
			break;
		case IDIVMOD:
			i = TOS();
			j = TOS2();
			if (0 == i) {
				Ip = GETW(1036);
				break;
			}
			k = j / i;
			TO_TOS2(k);
			k = j % i;
			TO_TOS(k);
			break;
		case IUDIVMD:
			u = TOS();
			v = TOS2();
			if (0 == u) {
				Ip = GETW(1036);
				break;
			}
			w = v / u;
			TO_TOS2(w);
			w = v % u;
			TO_TOS(w);
			break;
		case IAND:
			u = TOS() & TOS2();
			TO_TOS2(u);
			Sp += 4;
			break;
		case IOR:
			u = TOS() | TOS2();
			TO_TOS2(u);
			Sp += 4;
			break;
		case IXOR:
			u = TOS() ^ TOS2();
			TO_TOS2(u);
			Sp += 4;
			break;
		case ISHL:
			u = TOS2() << TOS();
			TO_TOS2(u);
			Sp += 4;
			break;
		case ISHR:
			u = TOS2() >> TOS();
			TO_TOS2(u);
			Sp += 4;
			break;
		case INEG:
			TO_TOS(-TOS());
			break;
		case IINV:
			TO_TOS(~TOS());
			break;
		case IINC:
			TO_TOS(TOS()+1);
			break;
		case IDEC:
			TO_TOS(TOS()-1);
			break;
		case INOT:
			i = TOS();
			i = (0 == i? -1: 0);
			TO_TOS(i);
			break;
		case IADDTO:
			u = pop();
			i = pop();
			j = GETW(u);
			j += i;
			PUTW(u, j);
			break;
		case IBSZ:
			u = TOS();
			for (i=0; u&1; i++)
				u >>= 1;
			TO_TOS(i);
			break;
		case ILESS:
			i = TOS2();
			i = i < pop()? -1: 0;
			TO_TOS(i);
			break;
		case IULESS:
			u = TOS2();
			v = pop();
			i = u < v? -1: 0;
			TO_TOS(i);
			break;
		case IGRTR:
			i = TOS2();
			i = i > pop()? -1: 0;
			TO_TOS(i);
			break;
		case IUGRTR:
			u = TOS2();
			v = pop();
			i = u > v? -1: 0;
			TO_TOS(i);
			break;
		case IEQUAL:
			i = TOS2();
			i = pop() == i? -1: 0;
			TO_TOS(i);
			break;
		case ICMOVE:
			w = pop();
			v = pop();
			u = pop();
			memmove(&Mem[v], &Mem[u], w);
			break;
		case ICSCAN:
			j = pop();
			w = pop();
			u = TOS();
			TO_TOS(0);
			for (i=w; 0 != i; i--) {
				if (Mem[u++] == j) {
					TO_TOS(u-1);
					break;
				}
			}
			break;
		case IDTEXT:
			w = pop();
			v = pop();
			u = TOS();
			i = memcmp(&Mem[u], &Mem[v], w);
			TO_TOS(i);
			break;
		case IFILL:
			j = pop();
			w = pop();
			u = pop();
			memset(&Mem[u], j, w);
			break;
		case ILOOKUP:
			lookup();
			break;
		case IKBRDY:
			nodelay(stdscr, 1);
			j = getch();
			if (j >= 0)
				ungetch(j);
			nodelay(stdscr, 0);
			push(j >= 0? -1: 0);
			break;
		case IKEY:
			push(getkey());
			break;
		case IEMIT:
			addch(pop()); refresh();
			break;
		case ICURSOR:
			j = pop();
			i = pop();
			move(j, i);
			break;
		case ICOLOR:
			u = pop();
			u = ((u & 0x70) >> 1) | (u & 0x07);
			color_set(u, NULL);
			color = u;
			break;
		case ICPOS:
			getyx(stdscr, j, i);
			push(j);
			push(i);
			break;
		case ICLREOL:
			clrtoeol();
			break;
		case ICLRSCR:
			clear();
			break;
		case IDSKIO:
			disk_io();
			break;
		default:
			printregs("illegal instruction");
			quit("halted");
			break;
		}
	}
}

int main(void) {
	init();
	load();
	run();
	fini();
	return 0;
}
